@@ -18,13 +18,12 @@
 
 
 #include <assert.h>
-#include <setjmp.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <string.h>
 
 #if !defined(Py_LIMITED_API)
-#define Py_LIMITED_API      0x03030000
+#define Py_LIMITED_API      0x03070000
 #endif
 
 #include <Python.h>
@@ -37,24 +36,16 @@ unsigned sipVersion;
 const char *sipVersionStr;
 unsigned abiVersion;
 char *sipName;
-stringList *includeDirList;
 
 /* Support for fatal error handling. */
-#define NO_EXCEPTION        0           /* No exception has been set. */
-#define EXCEPTION_SET       1           /* An exception has been set. */
-#define EXCEPTION_NEEDED    2           /* An exception needs to be set. */
-
 static char error_text[1000];
-static jmp_buf on_fatal_error;
 static PyObject *exception_type;
-static void raise_exception(int action);
+static void raise_exception(void);
 
 /* Forward declarations. */
 static PyObject *py_set_globals(PyObject *self, PyObject *args);
-static PyObject *py_parse(PyObject *self, PyObject *args);
+static PyObject *py_py2c(PyObject *self, PyObject *args);
 static PyObject *py_generateCode(PyObject *self, PyObject *args);
-static PyObject *py_generateExtracts(PyObject *self, PyObject *args);
-static PyObject *py_generateAPI(PyObject *self, PyObject *args);
 static PyObject *py_generateXML(PyObject *self, PyObject *args);
 static PyObject *py_generateTypeHints(PyObject *self, PyObject *args);
 
@@ -63,7 +54,6 @@ static int sipSpec_convertor(PyObject *obj, sipSpec **ptp);
 static int stringList_convertor(PyObject *obj, stringList **slp);
 static PyObject *stringList_convert_from(stringList *sl);
 static int extend_stringList(stringList **slp, PyObject *py_list, int no_dups);
-static void exception_set(void);
 
 
 /*
@@ -73,10 +63,8 @@ PyMODINIT_FUNC PyInit_code_generator(void)
 {
     static PyMethodDef methods[] = {
         {"set_globals", py_set_globals, METH_VARARGS, NULL},
-        {"parse", py_parse, METH_VARARGS, NULL},
+        {"py2c", py_py2c, METH_VARARGS, NULL},
         {"generateCode", py_generateCode, METH_VARARGS, NULL},
-        {"generateExtracts", py_generateExtracts, METH_VARARGS, NULL},
-        {"generateAPI", py_generateAPI, METH_VARARGS, NULL},
         {"generateXML", py_generateXML, METH_VARARGS, NULL},
         {"generateTypeHints", py_generateTypeHints, METH_VARARGS, NULL},
         {NULL, NULL, 0, NULL},
@@ -105,14 +93,13 @@ static PyObject *py_set_globals(PyObject *self, PyObject *args)
 {
     unsigned abi_major, abi_minor;
 
-    if (!PyArg_ParseTuple(args, "IzIIzOO&",
+    if (!PyArg_ParseTuple(args, "IzIIzO",
             &sipVersion,
             &sipVersionStr,
             &abi_major,
             &abi_minor,
             &sipName,
-            &exception_type,
-            stringList_convertor, &includeDirList))
+            &exception_type))
         return NULL;
 
     abiVersion = (abi_major << 8) | abi_minor;
@@ -127,56 +114,20 @@ static PyObject *py_set_globals(PyObject *self, PyObject *args)
 
 
 /*
- * Wrapper around parse().
+ * Wrapper around py2c().
  */
-static PyObject *py_parse(PyObject *self, PyObject *args)
+static PyObject *py_py2c(PyObject *self, PyObject *args)
 {
-    sipSpec *pt;
-    FILE *file;
-    char *filename;
-    stringList *versions, *backstops, *xfeatures, *sip_files;
-    int strict, protHack, action;
+    PyObject *spec;
+    const char *encoding;
+    extern sipSpec *py2c(PyObject *, const char *);
 
-    if (!PyArg_ParseTuple(args, "O&pO&O&O&p",
-            fs_convertor, &filename,
-            &strict,
-            stringList_convertor, &versions,
-            stringList_convertor, &backstops,
-            stringList_convertor, &xfeatures,
-            &protHack))
+    if (!PyArg_ParseTuple(args, "Os",
+            &spec,
+            &encoding))
         return NULL;
 
-    pt = sipMalloc(sizeof (sipSpec));
-
-    if (filename != NULL)
-    {
-        file = NULL;
-    }
-    else
-    {
-        file = stdin;
-        filename = "stdin";
-    }
-
-    if ((action = setjmp(on_fatal_error)) != NO_EXCEPTION)
-    {
-        raise_exception(action);
-        return NULL;
-    }
-
-    sip_files = NULL;
-
-    parse(pt, file, filename, strict, &versions, backstops, &xfeatures,
-            protHack, &sip_files);
-
-    transform(pt, strict);
-
-    return Py_BuildValue("(NsiNNN)", PyCapsule_New(pt, NULL, NULL),
-            pt->module->fullname->text,
-            (isComposite(pt->module) || useLimitedAPI(pt->module)),
-            stringList_convert_from(sip_files),
-            stringList_convert_from(versions),
-            stringList_convert_from(xfeatures));
+    return PyCapsule_New(py2c(spec, encoding), NULL, NULL);
 }
 
 
@@ -188,7 +139,7 @@ static PyObject *py_generateCode(PyObject *self, PyObject *args)
     sipSpec *pt;
     char *codeDir, *srcSuffix;
     const char *api_header;
-    int exceptions, tracing, releaseGIL, parts, docs, py_debug, action;
+    int exceptions, tracing, releaseGIL, parts, docs, py_debug;
     stringList *versions, *xfeatures, *sources;
 
     if (!PyArg_ParseTuple(args, "O&O&O&pppiO&O&pp",
@@ -205,72 +156,20 @@ static PyObject *py_generateCode(PyObject *self, PyObject *args)
             &py_debug))
         return NULL;
 
-    if ((action = setjmp(on_fatal_error)) != NO_EXCEPTION)
-    {
-        raise_exception(action);
-        return NULL;
-    }
-
     sources = generateCode(pt, codeDir, srcSuffix, exceptions, tracing,
             releaseGIL, parts, versions, xfeatures, docs, py_debug,
             &api_header);
 
+    if (sources == NULL)
+    {
+        raise_exception();
+        return NULL;
+    }
+
     return Py_BuildValue("(sN)", api_header, stringList_convert_from(sources));
 }
 
 
-/*
- * Wrapper around generateExtracts().
- */
-static PyObject *py_generateExtracts(PyObject *self, PyObject *args)
-{
-    sipSpec *pt;
-    stringList *extracts;
-    int action;
-
-    if (!PyArg_ParseTuple(args, "O&O&",
-            sipSpec_convertor, &pt,
-            stringList_convertor, &extracts))
-        return NULL;
-
-    if ((action = setjmp(on_fatal_error)) != NO_EXCEPTION)
-    {
-        raise_exception(action);
-        return NULL;
-    }
-
-    generateExtracts(pt, extracts);
-
-    Py_RETURN_NONE;
-}
-
-
-/*
- * Wrapper around generateAPI().
- */
-static PyObject *py_generateAPI(PyObject *self, PyObject *args)
-{
-    sipSpec *pt;
-    char *apiFile;
-    int action;
-
-    if (!PyArg_ParseTuple(args, "O&O&",
-            sipSpec_convertor, &pt,
-            fs_convertor, &apiFile))
-        return NULL;
-
-    if ((action = setjmp(on_fatal_error)) != NO_EXCEPTION)
-    {
-        raise_exception(action);
-        return NULL;
-    }
-
-    generateAPI(pt, pt->module, apiFile);
-
-    Py_RETURN_NONE;
-}
-
-
 /*
  * Wrapper around generateXML().
  */
@@ -278,21 +177,18 @@ static PyObject *py_generateXML(PyObject *self, PyObject *args)
 {
     sipSpec *pt;
     char *xmlFile;
-    int action;
 
     if (!PyArg_ParseTuple(args, "O&O&",
             sipSpec_convertor, &pt,
             fs_convertor, &xmlFile))
         return NULL;
 
-    if ((action = setjmp(on_fatal_error)) != NO_EXCEPTION)
+    if (generateXML(pt, pt->module, xmlFile) < 0)
     {
-        raise_exception(action);
+        raise_exception();
         return NULL;
     }
 
-    generateXML(pt, pt->module, xmlFile);
-
     Py_RETURN_NONE;
 }
 
@@ -304,21 +200,18 @@ static PyObject *py_generateTypeHints(PyObject *self, PyObject *args)
 {
     sipSpec *pt;
     char *pyiFile;
-    int action;
 
     if (!PyArg_ParseTuple(args, "O&O&",
             sipSpec_convertor, &pt,
             fs_convertor, &pyiFile))
         return NULL;
 
-    if ((action = setjmp(on_fatal_error)) != NO_EXCEPTION)
+    if (generateTypeHints(pt, pt->module, pyiFile) < 0)
     {
-        raise_exception(action);
+        raise_exception();
         return NULL;
     }
 
-    generateTypeHints(pt, pt->module, pyiFile);
-
     Py_RETURN_NONE;
 }
 
@@ -468,10 +361,9 @@ static int extend_stringList(stringList **slp, PyObject *py_list, int no_dups)
 
 
 /*
- * Display a one line error message describing a fatal error.  This does not
- * return.
+ * Display a one line error message describing a fatal error.
  */
-void fatal(const char *fmt, ...)
+int error(const char *fmt, ...)
 {
     va_list ap;
     size_t used = strlen(error_text);
@@ -481,15 +373,14 @@ void fatal(const char *fmt, ...)
     vsnprintf(&error_text[used], room, fmt, ap);
     va_end(ap);
 
-    /* Raise an exception. */
-    longjmp(on_fatal_error, EXCEPTION_NEEDED);
+    return -1;
 }
 
 
 /*
  * Append to the current error message.
  */
-void fatalAppend(const char *fmt, ...)
+void errorAppend(const char *fmt, ...)
 {
     va_list ap;
     size_t used = strlen(error_text);
@@ -504,10 +395,9 @@ void fatalAppend(const char *fmt, ...)
 /*
  * Raise an exception if needed.
  */
-static void raise_exception(int action)
+static void raise_exception()
 {
-    if (action == EXCEPTION_NEEDED)
-        PyErr_SetString(exception_type, error_text);
+    PyErr_SetString(exception_type, error_text);
 
     /*
      * The error text buffer may be used more than once as the Python exception
@@ -515,100 +405,3 @@ static void raise_exception(int action)
      */
     error_text[0] = '\0';
 }
-
-
-/*
- * Return to the Python interpreter after an exception has been set.
- */
-static void exception_set(void)
-{
-    longjmp(on_fatal_error, EXCEPTION_SET);
-}
-
-
-/*
- * Display a warning message.
- */
-void warning(Warning w, const char *fmt, ...)
-{
-    static char warning_text[1000];
-
-    va_list ap;
-    size_t used = strlen(warning_text);
-    size_t room = sizeof (warning_text) - used - 1;
-
-    va_start(ap, fmt);
-    vsnprintf(&warning_text[used], room, fmt, ap);
-    va_end(ap);
-
-    if (strchr(fmt, '\n') != NULL)
-    {
-        int ret;
-
-        ret = PyErr_WarnEx(
-                (w == DeprecationWarning ? PyExc_FutureWarning :
-                        PyExc_UserWarning),
-                warning_text, 1);
-
-        warning_text[0] = '\0';
-
-        if (ret < 0)
-            exception_set();
-    }
-}
-
-
-/*
- * Get the configuration of a set of bindings and update the list of tags and
- * disabled features.
- */
-void get_bindings_configuration(const char *sip_file, stringList **tags,
-        stringList **disabled)
-{
-    static PyObject *get_bindings_configuration = NULL;
-
-    PyObject *res;
-
-    /* Get the Python helper. */
-    if (get_bindings_configuration == NULL)
-    {
-        PyObject *mod;
-
-        if ((mod = PyImport_ImportModule("sipbuild.helpers")) == NULL)
-            exception_set();
-
-        get_bindings_configuration = PyObject_GetAttrString(mod,
-                "get_bindings_configuration");
-        Py_DECREF(mod);
-
-        if (get_bindings_configuration == NULL)
-            exception_set();
-    }
-
-    /* Call the helper. */
-    res = PyObject_CallFunction(get_bindings_configuration, "IsN",
-            abiVersion >> 8, sip_file,
-            stringList_convert_from(includeDirList));
-
-    if (res == NULL)
-        exception_set();
-
-    /* The result should be a 2-tuple of lists of strings. */
-    assert(PyTuple_Check(res));
-    assert(PyTuple_Size(res) == 2);
-
-    if (!extend_stringList(tags, PyTuple_GetItem(res, 0), 1))
-    {
-        Py_DECREF(res);
-        exception_set();
-    }
-
-    if (!extend_stringList(disabled, PyTuple_GetItem(res, 1), 1))
-    {
-        Py_DECREF(res);
-        exception_set();
-    }
-
-    /* Tidy up. */
-    Py_DECREF(res);
-}
